/*
 * BDT-Docs-Undo.c   modul s funkcemi pro Undo
 */

#define BDT_UNDO_C

#include "wed_editor.h"



// UNDO FUNKCE //

// pomocne funkce pro Undo
// uvolneni pameti, ktera je zabrana informacemi o Undo
// removes the most recent undo data
// return the remaining number of undo steps
ULONG FreeUndoData(TEditFile *teditfile)
//>>
{
     EditFile *doc = teditfile->GetEditFile();
     struct Undo *pundo;
//     UWORD undo_comm = doc->undo_info.undo_command;

     if (doc == NULL) 
          return 0;
     if (doc->undo_steps == 0)
          return 0;

     doc->undo_steps--;  // one step less
     pundo = doc->undo_info + doc->undo_steps;
     
     if (pundo->buf)
          LibFreePooled(GETPOOLHEADER, pundo->buf, pundo->bufsize);

     clrmem(pundo, sizeof(struct Undo));
     pundo->undo_command = UNDO_Empty;

     return doc->undo_steps;
}
//<<

// removes the oldest undo information and adds the new undo step
// return the remaining number of undo steps or 0 (error)
ULONG AddUndoData(TEditFile *teditfile, struct Undo *pundo)
//>>
{
     EditFile *doc = teditfile->GetEditFile();
     struct Undo *pundo2;
//     UWORD undo_comm = doc->undo_info.undo_command;

     if (doc == NULL) 
          return 0;

     if (doc->undo_steps < NUM_UNDO_STEPS)
     {
          doc->undo_info[doc->undo_steps++] = *pundo;
          return doc->undo_steps;
     }

     // remove the oldest undo 
     pundo2 = doc->undo_info;
     if (pundo2->buf)
          LibFreePooled(GETPOOLHEADER, pundo2->buf, pundo2->bufsize);
     memmove(doc->undo_info, doc->undo_info + 1, (NUM_UNDO_STEPS - 1) * sizeof(struct Undo));

     // add the new one
     doc->undo_info[(doc->undo_steps = NUM_UNDO_STEPS) - 1] = *pundo;

     return doc->undo_steps;
}
//<<

// frees all undo information for the document
void FreeAllUndoData(TEditFile *teditfile)
{
     EditFile *doc = teditfile->GetEditFile();
     
     while (FreeUndoData(teditfile) > 0) 
          ;

     doc->undo_steps = 0;
}

// inicializace undo
void InitUndoInfo(TEditFile *teditfile)
//>>
{
     EditFile *doc = teditfile->GetEditFile();

     if (doc == NULL) return;

     clrmem(doc->undo_info, NUM_UNDO_STEPS * sizeof(struct Undo));
     doc->undo_info[0].undo_command = UNDO_Empty;

     doc->undo_steps = 0;
}
//<<

// ukonceni undo + uvolneni vseho
void FreeUndoInfo(TEditFile *teditfile)
//>>
{
     if (teditfile == NULL) return;

     FreeAllUndoData(teditfile);
}
//<<

// return the number of the given paragraph in the document
// return 0 if para is not found
ULONG GetParaNumber(TEditFile *teditfile, Paragraph *para)
{
     EditFile *doc = teditfile->GetEditFile();
     Paragraph *p;
     ULONG num;

     for(num = 1, p = doc->para_first; p && (p != para); p = p->next)
          num++;
     if (p == NULL) 
          num = 0;

     return num;
}
     
#define CURSORLEAVEPOS_INDICATOR   0x80000000     // je-li tento bit nastaven v data - posunul se kurzor

// aktualizace undo info po nove operaci
// predpoklada se, ze data z predchozi operace jsou jiz uvonena;
// a nova data jsou jiz ulozena / popr. viz data /
// jinak by v pameti byly dve undo data najednou
// vyjimkou je samozrejme UNDO_Empty
// tato funkce spravuje take zvyraznovani menu Undo
//
// undo update is needed actually for setting para_after parts of undo info
//   it should be called after the operation is finished
//   also, with new_undo = UNDO_Empty, the undo is cleared (e.g., after open or so)
void UpdateUndoInfo(TEditFile *teditfile, UWORD new_undo, ULONG data)
//>>
{
     EditFile *doc = teditfile->GetEditFile();
     struct Undo *pundo; 
     
     // if we undo things, we do not store actions to undo memory
     if (teditfile->UndoingActive()) return;
     
     if (doc == NULL) return;
     if (new_undo == UNDO_Special) return;
     // return if no undo is available
     if (doc->undo_steps == 0) return;
     pundo = &doc->undo_info[doc->undo_steps - 1];

  // nic se nestalo - zrusi se undo informace
     switch (new_undo)
     {
          case UNDO_Empty:    // clear undo info
               FreeAllUndoData(teditfile);
               break;

          default:
               // set the cursor position after the action
               pundo->para_after = GetParaNumber(teditfile, doc->cursor.para);
               pundo->charoff_after = doc->cursor.chars_off;
               break;
     }
}
//<<

// funkce, ktere ulozi = uschova data pro undo
// pokud je treba dosadit neco do data pro UpdateUndoInfo,
// vrati to tato funkce
// new_undo urcuje typ akce; podle prechoazejiciho unda musi byt vymazano
// to co predchazelo
// data1, data2 = potrebne informace, jsou-li treba
// nektere operace nemuseji tuto funkci vubec volat
// musi si nutne kontrolovat, zda je undo aktivni - nekotroluje to volajici funkce!
// pokud dany prikaz nepodporuje undo - musi se vymazat existujici informace
ULONG SaveUndoInfo(TEditFile *teditfile, UWORD new_undo, ULONG data1, ULONG data2, ULONG data3)
//>>
{
     EditFile *doc = teditfile->GetEditFile();

     struct Undo undo;
     Location pom1, pom2;
     ULONG data = 0;
     BOOL error = FALSE;

     // necessary to set buf, bufsize to NULL, 0
     clrmem(&undo, sizeof(struct Undo));

     // if we undo things, we do not store actions to undo memory
     if (teditfile->UndoingActive()) return 0;
     
     if (doc == NULL) 
          return 0;
     if (new_undo == UNDO_Special) 
          return 0;

     switch (new_undo)
     {
          case UNDO_FontFormat:    // needs update
               data1 = data2 = 0;
               if (NOT doc->block)
               {
                    undo.undo_command = UNDO_FontFormat;
                    undo.bufsize = 4;
                    undo.buf = (STRPTR) LibAllocPooled(GETPOOLHEADER, undo.bufsize);

                    if (undo.buf)
                    {
                         undo.buf[0] = *GetPosMB(doc->cursor.block, doc->cursor.block_off);
                         undo.buf[1] = '\0';
                    }
                    else error = TRUE;
                    
                    // para_beg,end are set to current position, update change end to the right position
                    undo.para_beg = undo.para_end = undo.para_before = undo.para_after = GetParaNumber(teditfile, doc->cursor.para);
                    undo.charoff_beg = undo.charoff_end = undo.charoff_before = undo.charoff_after = doc->cursor.chars_off;
                    undo.block = FALSE;

                    data = 1;
                    break;
               }

          case UNDO_DelText:  // data1, data2 jsou ukazatele na Location pocatku a konce
                              // data3 points to the original cursor position
               if (NOT doc->block)
               {
                    Location *l1 = (Location *) data1;
                    Location *l2 = (Location *) data2;
                    Location *l3 = (Location *) data3;

                    undo.para_beg = GetParaNumber(teditfile, l1->para);
                    undo.charoff_beg = l1->chars_off;
                    undo.para_end = GetParaNumber(teditfile, l2->para);
                    undo.charoff_end = l2->chars_off;

                    if (undo.para_beg && undo.para_end && l1 && l2)
                    {
                         // we have to set the block
                         pom1 = doc->block_begin;
                         pom2 = doc->block_end;
                         doc->block_begin = *l1;
                         doc->block_end = *l2;
                         (void) GetNextPos(teditfile, &doc->block_end, TRUE, 1);
                         doc->block = TRUE;

                         // get block size
                         undo.bufsize = GetBlockSize(teditfile, TRUE) + 4; // 4 - reserve for one CH_EOL and final '\0'
                         undo.buf = (STRPTR) LibAllocPooled(GETPOOLHEADER, undo.bufsize);

                         if (undo.buf)
                         {    // here we rely on clearing memory by Alloc ^^^ - thus, string is created and clrmem is not needed
                              // save the block
                              if (SaveClipASCII(teditfile, (CLIPHANDLE) undo.buf, FALSE))
                              {
                                   LibFreePooled(GETPOOLHEADER, undo.buf, undo.bufsize);
                                   BeepEditor();
                                   error = TRUE;
                              }
                         }
                         else error = TRUE;

                         // restore block information
                         doc->block = FALSE;
                         doc->block_begin = pom1;
                         doc->block_end = pom2;
                         data = 0;

                         // set pointers
                         undo.undo_command = new_undo;
                         // buf and bufsize were set before
                         undo.para_before = undo.para_after = GetParaNumber(teditfile, doc->cursor.para);
                         undo.charoff_before = undo.charoff_after = doc->cursor.chars_off;
                         if (l3)	// set correct cursor before, if it is supplied
                         {
                              undo.para_before = GetParaNumber(teditfile, l3->para);
                              undo.charoff_before = l3->chars_off;
                         }
                         undo.block = FALSE;
                    }
                    else error = TRUE;
                    break;
               }
               else
               {
                    data1 = data2 = 0;
               }

          case UNDO_DelBlock: // zde data1 / 2 obsahuji pocet vkladanych znaku - ignored for now
               if (NOT doc->block)
               {
                    error = TRUE;
                    break;
               }

               undo.para_beg = GetParaNumber(teditfile, doc->block_begin.para);
               undo.charoff_beg = doc->block_begin.chars_off;
               undo.para_end = GetParaNumber(teditfile, doc->block_end.para);
               undo.charoff_end = doc->block_end.chars_off;

               if (undo.para_beg && undo.para_end)
               {
                    undo.bufsize = GetBlockSize(teditfile, TRUE) + 4; // 4 - reserve for one CH_EOL and final '\0'
                    undo.buf = (STRPTR) LibAllocPooled(GETPOOLHEADER, undo.bufsize);

                    if (undo.buf)
                    {    // here we rely on clearing memory by Alloc ^^^ - thus, string is created even without clrmem
                         // block is selected!
                         if (SaveClipASCII(teditfile, (CLIPHANDLE) undo.buf, FALSE))
                         {
                              LibFreePooled(GETPOOLHEADER, undo.buf, undo.bufsize);
                              BeepEditor();
                              error = TRUE;
                         }
                    }
                    else error = TRUE;

                    data = 0;

                    // set pointers
                    undo.undo_command = new_undo;
                    // buf and bufsize were set before
                    undo.para_before = undo.para_after = GetParaNumber(teditfile, doc->cursor.para);
                    undo.charoff_before = undo.charoff_after = doc->cursor.chars_off;
                    undo.block = TRUE;
               }
               else error = TRUE;
               break;

          case UNDO_InsText:  // zde data1 obsahuji pocet vkladanych znaku (nebo INSPARA pro odstavec)
                              // needs update
               undo.undo_command = UNDO_InsText;
               undo.buf = NULL;
               undo.bufsize = 0;
               
               // para_beg,end are set to current position, update change end to the right position
               undo.para_beg = undo.para_end = undo.para_before = undo.para_after = GetParaNumber(teditfile, doc->cursor.para);
               undo.charoff_beg = undo.charoff_end = undo.charoff_before = undo.charoff_after = doc->cursor.chars_off;
               undo.block = FALSE;
                    
               data = data1;
               break;
               
          case UNDO_OverwriteText: // currently just for one character
                                   // needs update
               undo.undo_command = UNDO_OverwriteText;
               undo.bufsize = 4;
               undo.buf = (STRPTR) LibAllocPooled(GETPOOLHEADER, undo.bufsize);

               if (undo.buf)
               {
                    undo.buf[0] = *GetPosMB(doc->cursor.block, doc->cursor.block_off);
                    undo.buf[1] = '\0';
               }
               else error = TRUE;
               
               // para_beg,end are set to current position, update change end to the right position
               undo.para_beg = undo.para_end = undo.para_before = undo.para_after = GetParaNumber(teditfile, doc->cursor.para);
               undo.charoff_beg = undo.charoff_end = undo.charoff_before = undo.charoff_after = doc->cursor.chars_off;
               undo.block = FALSE;
                    
               data = 1;
               break;
               
          case UNDO_InsClip: // data1, data2 jsou ukazatele na Location pocatku a konce
               {
                    Location *l1 = (Location *) data1;
                    Location *l2 = (Location *) data2;

                    if (l1 && l2)
                    {
                         undo.undo_command = UNDO_InsClip;
                         undo.buf = NULL;
                         undo.bufsize = 0;
                         
                         // para_beg,end are set to current position, update change end to the right position
                         undo.para_beg = GetParaNumber(teditfile, l1->para);
                         undo.charoff_beg = l1->chars_off;
                         undo.para_end = GetParaNumber(teditfile, l2->para);
                         undo.charoff_end = l2->chars_off;
                         undo.para_before = undo.para_after = GetParaNumber(teditfile, doc->cursor.para);
                         undo.charoff_before = undo.charoff_after = doc->cursor.chars_off;
                         undo.block = (BOOL) data3;
                    }
                    else error = TRUE;
                         
                    data = 0;
               }
               break;
               
          case UNDO_ReplaceText:  // data1, data2 jsou ukazatele na Location pocatku a konce
               {
                    Location *l1 = (Location *) data1;
                    Location *l2 = (Location *) data2;
                    BOOL block;
                    
                    undo.para_beg = GetParaNumber(teditfile, l1->para);
                    undo.charoff_beg = l1->chars_off;
                    undo.para_end = GetParaNumber(teditfile, l2->para);
                    undo.charoff_end = l2->chars_off;

                    if (undo.para_beg && undo.para_end && l1 && l2)
                    {
                         // we have to set the block
                         pom1 = doc->block_begin;
                         pom2 = doc->block_end;
                         block = doc->block;

                         doc->block_begin = *l1;
                         doc->block_end = *l2;
                         (void) GetNextPos(teditfile, &doc->block_end, TRUE, 1);
                         doc->block = TRUE;

                         // get block size
                         undo.bufsize = GetBlockSize(teditfile, TRUE) + 4; // 4 - reserve for one CH_EOL and final '\0'
                         undo.buf = (STRPTR) LibAllocPooled(GETPOOLHEADER, undo.bufsize);

                         if (undo.buf)
                         {    // here we rely on clearing memory by Alloc ^^^ - thus, string is created and clrmem is not needed
                              // save the block
                              if (SaveClipASCII(teditfile, (CLIPHANDLE) undo.buf, FALSE))
                              {
                                   LibFreePooled(GETPOOLHEADER, undo.buf, undo.bufsize);
                                   BeepEditor();
                                   error = TRUE;
                              }
                         }
                         else error = TRUE;

                         // restore block information
                         doc->block = block;
                         doc->block_begin = pom1;
                         doc->block_end = pom2;
                         data = 0;

                         // set pointers
                         undo.undo_command = new_undo;
                         // buf and bufsize were set before
                         undo.para_before = undo.para_after = GetParaNumber(teditfile, doc->cursor.para);
                         undo.charoff_before = undo.charoff_after = doc->cursor.chars_off;
                         undo.block = FALSE;
                    }
                    else error = TRUE;
               }
               break;
               
          case UNDO_Empty:
               error = TRUE;  // to clear the undo
               break;
     }

  // if there is an error, free all undo
     if (error)
          FreeAllUndoData(teditfile);
     else // make the undo information active
          (void) AddUndoData(teditfile, &undo);
          
  // vraceni dat
     return(data);
}
//<<

