/*
 * BDT-Docs-Undo.c   modul s funkcemi pro Undo
 */

#define BDT_UNDO_C

#include "wed_editor.h"

// notes:
//   - do explicitily (or note somewhere) that undo (correct buf structure as string)
//     relies on the automatic zeroing (cleaning) of the allocated memory
//   - add to dispatch: after every call of SaveUndoInfo, UpdateUndoInfo must be called
// to be added:
//   - BOOL undoing_active as private, init to FALSE, + function
// undo functions (save + update undo info + ...) should not do anything 
// during the process of undoing
//   - goto column functions

BOOL TEditFile::UndoingActive(void)
{
     return undoing_active;
}

void TEditFile::CmGotoColumn()
{
     if (DispatchMessages(this, MN_GoToColumn, 0, NULL, FALSE))
          updatedisplay();
}

void TEditFile::CmGotoGivenColumn(ULONG col)
{
     if (DispatchMessages(this, MN_GoToColumn, 0, &col, FALSE))
          updatedisplay();
}

// funkce, ktera provede vlastni undo akci
// musi si zkontrolovat, jsou-li undo data a je-li undo aktivni
// this function is not called from Dispatch (ineffective), it rather does
// the job by calling different DispatchMessages functions
void TEditFile::DoUndoOperation(void)
//>>
{
     EditFile *doc = &editfile;

     Undo *pundo;
     UWORD undo_command;
     ULONG i, pom;
     BOOL draw = FALSE;
     UBYTE *pchar;

     // return if no undo is available
     if (doc->undo_steps == 0) return;
     
     pundo = &doc->undo_info[doc->undo_steps - 1];
     undo_command = pundo->undo_command;
     if ((undo_command == UNDO_Empty) || (undo_command == UNDO_Special)) 
          return;

     // wait pointer
     undoing_active = TRUE;
     ChangePointer(this, NULL, TRUE, TRUE);

     // set cursor position - position after the action
     pom = pundo->para_after;
     if (DispatchMessages(this, MN_GoToLine, 0, &pom, FALSE))
          draw = TRUE;
     pom = pundo->charoff_after + 1;
     if (DispatchMessages(this, MN_GoToColumn, 0, &pom, FALSE))
          draw = TRUE;
     
     // undo 
     switch (undo_command)
     {
          case UNDO_DelBlock:
          case UNDO_DelText:  // inserted the deleted text
               if ((pundo->bufsize == 0) || (pundo->buf == NULL))
                    break;
               
               if (DispatchMessages(this, MN_Paste, 0, pundo->buf, FALSE))
                    draw = TRUE;
               break;

          case UNDO_FontFormat:
               if ((pundo->bufsize == 0) || (pundo->buf == NULL))
                    break;

               // select the original block
               pom = pundo->para_beg;
               if (DispatchMessages(this, MN_GoToLine, 0, &pom, FALSE))
                    draw = TRUE;
               pom = pundo->charoff_beg + 1;
               if (DispatchMessages(this, MN_GoToColumn, 0, &pom, FALSE))
                    draw = TRUE;
               
               pom = pundo->para_end;
               if (DispatchMessages(this, MN_GoToLineBlock, 0, &pom, TRUE))
                    draw = TRUE;
               pom = pundo->charoff_end + 1;
               if (DispatchMessages(this, MN_GoToColumnBlock, 0, &pom, TRUE))
                    draw = TRUE;

               // paste over the saved version
               if (DispatchMessages(this, MN_Paste, 0, pundo->buf, FALSE))
                    draw = TRUE;
               break;
               
          case UNDO_InsText:
               // select the inserted text
               pom = pundo->para_before;
               if (DispatchMessages(this, MN_GoToLineBlock, 0, &pom, TRUE))
                    draw = TRUE;
               pom = pundo->charoff_before + 1;
               if (DispatchMessages(this, MN_GoToColumnBlock, 0, &pom, TRUE))
                    draw = TRUE;

               // deleted
               if (DispatchMessages(this, MN_Erase, 0, NULL, FALSE))
                    draw = TRUE;
               break;
               
          case UNDO_InsClip:
               // select the inserted text
               pom = pundo->para_beg;
               if (DispatchMessages(this, MN_GoToLine, 0, &pom, FALSE))
                    draw = TRUE;
               pom = pundo->charoff_beg + 1;
               if (DispatchMessages(this, MN_GoToColumn, 0, &pom, FALSE))
                    draw = TRUE;
               
               pom = pundo->para_end;
               if (DispatchMessages(this, MN_GoToLineBlock, 0, &pom, TRUE))
                    draw = TRUE;
               pom = pundo->charoff_end + 1;
               if (DispatchMessages(this, MN_GoToColumnBlock, 0, &pom, TRUE))
                    draw = TRUE;

               // deleted
               if (DispatchMessages(this, MN_Erase, 0, NULL, FALSE))
                    draw = TRUE;
               break;
               
          case UNDO_OverwriteText:
               if ((pundo->bufsize == 0) || (pundo->buf == NULL))
                    break;

               // set cursor to original position
               pom = pundo->para_before;
               if (DispatchMessages(this, MN_GoToLineBlock, 0, &pom, TRUE))
                    draw = TRUE;
               pom = pundo->charoff_before + 1;
               if (DispatchMessages(this, MN_GoToColumnBlock, 0, &pom, TRUE))
                    draw = TRUE;

               // put the original character on its place
               pchar = GetPosMB(doc->cursor.block, doc->cursor.block_off);
               if (isalpha(*pchar))
                    *pchar = *pundo->buf;
               break;

          case UNDO_ReplaceText:
               // select the inserted text
               pom = pundo->para_before;
               if (DispatchMessages(this, MN_GoToLineBlock, 0, &pom, TRUE))
                    draw = TRUE;
               pom = pundo->charoff_before + 1;
               if (DispatchMessages(this, MN_GoToColumnBlock, 0, &pom, TRUE))
                    draw = TRUE;

               // deleted
               if (DispatchMessages(this, MN_Erase, 0, NULL, FALSE))
                    draw = TRUE;

               // insert replaced text
               if ((pundo->bufsize == 0) || (pundo->buf == NULL))
                    break;
               
               if (DispatchMessages(this, MN_Paste, 0, pundo->buf, FALSE))
                    draw = TRUE;
               break;
     }

     // zruseni ukazatele
     teditfile->undoing_active = FALSE;
     ChangePointer(this, NULL, FALSE, FALSE);
     
     // remove the undo information
     (void) FreeUndoData(teditfile);

     if (draw)
          updatedisplay();
}
//<<

