/* $Header: /home/xplore/cvs/xpl4gr/xstring.hpp,v 1.4 2002/10/24 16:13:53 cizek Exp $ */

/*! \file xstring.hpp
 * header file for the class of the string database.
 */ 
#ifndef XPLSTRING
#define XPLSTRING

#include<stdlib.h>
#include<stdio.h>
#include<string.h>
#include<time.h>

#ifdef __BORLANDC__
#include<alloc.h>
#endif

#ifndef XPLSORT
#include "xplerror.h"
#endif

#include "global.h"
double x4atof (char *tok);  // aus initxa.hpp

/*! \brief Maximal number of strings
 * 
 * The minimal memory space allocated for the string database. 
 */
#define MAXSTRING   32768   

/*! \brief Added strings by expanding the database
 * 
 * Number of strings which will be added to the database if it should be expanded. 
 */
#define STRINGSPLUS  32768

/*! \brief Deleted strings when shrinking the database
 * 
 * Number of strings which will be expunged from the database when it becomes too large,
 * in particularly larger than $MAXSTRING$ + $STRINGSMINUS$. 
 */
#define STRINGSMINUS (2*STRINGSPLUS)

/*! \brief An element of the string database. 
 * 
 * Each element contains a pointer on the text string and a number of variables pointing on it. 
 */
struct XplStringElement
{
    /*! \brief String pointer.
     * 
     * Points on the text string of this element. 
     */
    char *txt;
    
    /*! \brief Number of links.
     * 
     * Stores number of variables pointing on this element. 
     */	
    unsigned long links;
};

/*! \brief A string database class. 
 * 
 * All strings are stored in one string database. The database is an array of $XplStringElement$
 * and an array of their indices used for sorting.
 * Although strings are internally numbered 0,1,2,..., interface behaves as if
 * they were numbered 1,2,3,... for compatibility with X4 
 */
class XplStringDatabase
{ 
    /*! \brief The string database.
     * 
     * The array of elements containing a string and an integer which stores number of variables pointing
     * to this string. The array expands and shrinks according to the number of strings. The strings are not sorted. 
     */
    XplStringElement *list;  
    
    /*! \brief Array of the indices of the strings.
     * 
     * This array is used for sorting and searching the database. 
     */
    unsigned long *indices;
    
    /*! \brief Used strings.
     * 
     * Number of string entries in the database including deleted strings also. It is the index of the last entry + 1. 
     */
    unsigned long used;
    
    /*! \brief First empty element.
     * 
     * First space ready for inserting a new element (arose after deleting a string). If there is no empty space in the array,
     * a new string will be inserted at the end of the array and $empty$ contains index $used$. 
     */
    unsigned long empty;
    
    /*! \brief Actual size of the database.
     * 
     *  Size of allocated memory for the database. Firstly equal to $MAXSTRING$. 
     */
    unsigned long size;
    
    /*! \brief Number of strings.
     * 
     *  Length of the array $indices$, i.e. it contains the actual number of nondeleted strings. 
     */
    unsigned long nstrings;
    
    /*! \brief Pointer for all deleted strings.
     * 
     *  A pointer for an universal string at which point all deleted strings. 
     */
    char *emptystring;

    /*! \brief Private method. It looks for a given string in the database.
     * 
     * \param newtxt a constant character pointer. String which is being searched.
     * \param mid an unsigned long pointer. Index of the string if found in the database. Index of the string which will be after $newtxt$ if not found in the database.
     * \return Search result. 1 if found, 0 else.
     * \sa XplStringDatabase(), ~XplStringDatabase()
     */
    int search (const char *newtxt, long *mid);

    public:

    /*! \brief The constructor.
     * 
     * Allocates memory for the database, clean it and sets up the variables. 
     */
    XplStringDatabase (void);
    
    /*! \brief The destructor. 
     *
     * Deletes the database.
     */
    ~XplStringDatabase ();
    
    /*! \brief Returns the number of occupied entries.
     * 
     * Returns the number of used elements of the database, however some of them may be already deleted. 
     * 
     * \return Entries result. Number of elements used.
     * \sa XplStringDatabase(), ~XplStringDatabase()
     */
    unsigned long entries (void);	

    /*! \brief Returns the number of strings.
     * 
     * Returns the number of strings in the database. 
     * 
     * \return Number of strings.
     * \sa XplStringDatabase(), ~XplStringDatabase()
     */
    unsigned long numstrings (void);	

    /*! \brief Inserts a string into the database.
     * 
     * Tries to insert $newtxt$ into the database and returns its position. 
     * If the string already exists, the position of the existing string will be returned and 
     * the internal counter will be increased. If the string is new, it will be added 
     * at the very first empty place of the database and the array of indices will be sorted immediately.
     * If there is not enough space in the database, it will be expanded (if possible).
     * 
     * Necessary to handle the errors when not enought memory!!!
     * 
     * \param newtxt a constant character pointer. String which is being inserted.
     * \return Insert result. Index where the string $newtxt$ was inserted.
     * \sa XplStringDatabase(), ~XplStringDatabase()
     */
    unsigned long insert (const char *newtxt);
    
    /*! \brief Adds existing string.
     *
     * Increases the internal counter for an existing string - it actually adds this string into the database.
     *
     * \param no an unsigned long argument. Index of the existing string being added into the database.
     * \sa XplStringDatabase(), ~XplStringDatabase()
     */
    void add (unsigned long no);
    
    /*! \brief Decrease the number of links pointing at the string with index $no$.
     * 
     * If there is no else variable pointing at this string, it will be removed from the database and
     * the according element of the array $list$ will point at $emptystring$. 
     * If $no$ was the last element, the possibility of shrinking the database will be checked.
     * 
     * \param no an unsigned long argument. Index of the string being removed from the database.
     * \sa XplStringDatabase(), ~XplStringDatabase()
     */    
    void kill (unsigned long no);
    
    /*! \brief Appends the stringbase to the file $filename$.
     * 
     * Writes the complete contents of the string database.
     * 
     * \param filename a constant character pointer. Name of the file for the output.
     * \sa XplStringDatabase(), ~XplStringDatabase(), sdump()
     */
    void dump (char *filename);
    
    /*! \brief Dumps the string database briefly into a file.
     * 
     * Writes down the array of $indices$ and the corresponding non-empty elements of $list$ array.
     * 
     * \param filename a constant character pointer. Name of the file for the output.
     * \param howtoopen a constant character pointer. Parameter how to open the file, the correctness 
     * is not checked. Default is "w".
     * \sa XplStringDatabase(), ~XplStringDatabase(), dump()
     */    
    void sdump (char *filename, char *howtoopen = NULL);
    
    /*!
     * \param no an unsigned long argument. Index of the string from the database.
     * \return character pointer.
     * \sa XplStringDatabase(), ~XplStringDatabase()
     */
    const char *operator[] (unsigned long no);
};

extern XplStringDatabase xplorebase;

class xplstring
{ unsigned strno;

  public:

  xplstring ();
  xplstring (const char *txt);
  xplstring (const xplstring &txt);
  ~xplstring ();
  void kill ();
  xplstring& operator= (const xplstring &left);
  void setnull (void);
  int xplisnull(void) const;
  unsigned xplstrlen (void) const;
  xplstring xplstrdup (void) const;
//  xplstring xplstrlwr (void);
//  xplstring xplstrupr (void);
//  xplstring xplstrrev (void);
  double xplatof (void) const;
  int xplatoi (void) const;
  long xplatol (void) const;
  double xplstrtod (void) const;
  xplstring xplstrchr (int ch) const;
  xplstring xplstrrchr (int ch) const;
//  xplstring xplstrset (int ch);
//  xplstring xplstrnset (int ch, unsigned n);
  long xplstrtol (int base) const;
  unsigned long xplstrtoul (int base) const;
//  xplstring xplitoa (int val, int base);
//  xplstring xplltoa (long val, int base);
//  xplstring xplultoa (unsigned long val, int base);
  float xplstrtof (void) const;
  xplstring xplstrcpy (const xplstring txt) const;
  xplstring xplstrcat (const xplstring txt) const;
  int xplstrcmp (const xplstring txt) const;
  unsigned xplstrcspn (const xplstring txt) const; 
  xplstring xplstrpbrk (const xplstring txt) const;
  unsigned xplstrspn (const xplstring txt) const; 
  xplstring xplstrstr (const xplstring txt) const;
  xplstring xplstrtok (const xplstring txt) const;
  xplstring xplstrtoknull (const xplstring txt) const;
//  int xplstricmp (const xplstring txt);
  xplstring xplstrcpy (const char *txt) const;
  xplstring xplstrcat (const char *txt) const;
  int xplstrcmp (const char *txt) const;
  unsigned xplstrcspn (const char *txt) const;
  xplstring xplstrpbrk (const char *txt) const;
  unsigned xplstrspn (const char *txt) const;
  xplstring xplstrstr (const char *txt) const;
  xplstring xplstrtok (const char *txt) const;
  xplstring xplstrtoknull (const char *txt) const;
//  int xplstricmp (const char *txt);
  xplstring xplecvt (double val, int ndig, int *pdecpt, int *psign) const;
  xplstring xplfcvt (double val, int ndig, int *pdecpt, int *psign) const;
  xplstring xplstrncpy (const xplstring txt, unsigned n) const;
  xplstring xplstrncat (const xplstring txt, unsigned n) const;
  int xplstrncmp (const xplstring txt, unsigned n) const;
//  unsigned xplstrxfrm (const xplstring txt, unsigned n) const;
//  int xplstrnicmp (const xplstring txt, unsigned n);
  xplstring xplstrncpy (const char *txt, unsigned n) const;
  xplstring xplstrncat (const char *txt, unsigned n) const;
  int xplstrncmp (const char *txt, unsigned n) const;
//  unsigned xplstrxfrm (const char *txt, unsigned n) const;
  int operator< (const xplstring &r) const;
  int operator> (const xplstring &r) const;
  int operator== (const xplstring &r) const;
  int operator!= (const xplstring &r) const;
  xplstring operator+ (const xplstring &r) const;
  xplstring operator+ (long r) const;
  xplstring operator* (const xplstring &r) const;
  xplstring (double d);
  char* tostr (void) const;
  const char* tocstr (void) const;
//  xplstring gcvt (double val, int ndig, char *);
  void print () const;
};

#endif

